#ifndef __MESSAGE_H__
#define __MESSAGE_H__

#include <map>
#include <sstream>
#include <vector>

#include "ONSClient.h"

namespace ons {

class SystemPropKey {
   public:
    SystemPropKey() {}
    ~SystemPropKey() {}
    static const char* TAG;
    static const char* KEY;
    static const char* MSGID;
    static const char* RECONSUMETIMES;
    static const char* STARTDELIVERTIME;
};

class ONSCLIENT_API Message {
   public:
    Message();

    /**
     * Construct message with topic and body
     * @param topic Message topic, characters allowed are [a-zA-Z0-9], underline '_' and hyphen '-', total length
     * within 3~64 inclusive
     * @param body  Message body, maximum size is 4MB
     */
    Message(const std::string& topic, const std::string& body);

    Message(const std::string& topic, const std::string& tags, const std::string& body);

    /**
     * Construct message with given null terminated C-string
     * @param topic
     * @param tags Optional, may be null.
     * @param body
     */
    Message(const char* topic, const char* tags, const char* body);

    /**
     * Construct message with given pointer to C-string and length
     * @param topic
     * @param topic_size
     * @param tags
     * @param tags_length
     * @param body
     * @param body_length
     */
    Message(const char* topic, size_t topic_size, const char* tags, size_t tags_length, const char* body,
            size_t body_length);

    /**
     * Construct message with null terminated C-string
     * @param topic
     * @param tags
     * @param keys
     * @param body
     */
    Message(const char* topic, const char* tags, const char* keys, const char* body);

    virtual ~Message();

    Message(const Message& other);
    Message& operator=(const Message& other);

    // userProperties was used to save user specific parameters which doesn't
    // belong to SystemPropKey
    void putUserProperties(const char* key, const char* value);
    const char* getUserProperties(const char* key) const;
    void setUserProperties(std::map<std::string, std::string>& userProperty);
    std::map<std::string, std::string> getUserProperties() const;

    // systemProperties only save parameters defined in SystemPropKey, please do
    // not add other parameters into systemProperties, else it was not saved.
    void putSystemProperties(const char* key, const char* value);
    const char* getSystemProperties(const char* key) const;
    void setSystemProperties(std::map<std::string, std::string>& systemProperty);
    std::map<std::string, std::string> getSystemProperties() const;

    const char* getTopic() const;
    void setTopic(const char* topic);

    const char* getTag() const;
    void setTag(const char* tags);

    const char* getKey() const;
    void setKey(const char* keys);

    const char* getMsgID() const;
    void setMsgID(const char* msgId);

    long long getStartDeliverTime() const;

    void setStartDeliverTime(long long level);

    const char* getBody() const;

    const char* getByteBody(int* len) const;

    std::string getMsgBody() const;

    size_t getBodySize() const;

    void setMsgBody(const std::string& body);
    void setBody(unsigned char* body, int len);

    int getReconsumeTimes() const;
    void setReconsumeTimes(int reconsume_times);

    long long getStoreTimestamp() const;
    void setStoreTimestamp(long long store_timestamp);

    std::string toString() const;

    std::string toSystemString() const;

    std::string toUserString() const;

    long long getQueueOffset() const;
    void setQueueOffset(long long queueOffset);

   protected:
    void Init(const std::string& topic, const std::string& tags, const std::string& keys, const std::string& body);

   private:
    std::string topic_;
    std::string body_;
    long long store_timestamp_;
    long long queue_offset_;
    std::map<std::string, std::string> system_properties_;
    std::map<std::string, std::string> user_properties_;
};

}  // namespace ons
#endif
